import re, os

__all__=['acceptableSyntax','slimmer','css_slimmer',
         'html_slimmer','xhtml_slimmer']


# HTML
html_comments_oneline = re.compile(r'<!--.*?-->')
html_inline_css = re.compile(r'<style.*?>.*?</style>',
                             re.I|re.MULTILINE|re.DOTALL)
quoted_numerical_attribute = re.compile(r"(\S+=(\"|')\d+(\"|'))", re.MULTILINE|re.DOTALL)
any_tag = re.compile(r"<.*?>", re.I|re.MULTILINE|re.DOTALL)
excess_whitespace = re.compile(r' \s+|\s +', re.MULTILINE)

# CSS
css_comments = re.compile(r'/\*.*?\*/', re.MULTILINE|re.DOTALL)

hex_colour = re.compile(r'#\w{2}\w{2}\w{2}')

# Define the syntax options we accept
HTML = 'html'
XHTML = 'xhtml'
CSS = 'css'

OK_SYNTAX = (HTML, XHTML, CSS)

def acceptableSyntax(syntax):
    """ return the syntax as we recognize it or None """
    syntax = str(syntax).lower().strip().replace(' ','').replace('-','')
    syntax = syntax.replace('stylesheet','css') # allow for alias
    if syntax in OK_SYNTAX:
        return syntax
    else:
        return None

def slimmer(code, syntax=XHTML):
    """ wrap all function we have """
    if syntax == XHTML:
        return _xhtml_slimmer(code)
    elif syntax == HTML:
        return _html_slimmer(code)
    elif syntax == CSS:
        return _css_slimmer(code)
    
def _css_slimmer(css):
    """ remove repeating whitespace ( \t\n) """

    css = css_comments.sub('', css) # remove comments
    css = re.sub(r'\s\s+', '', css) # >= 2 whitespace becomes no whitespace
    css = re.sub(r'\s+{','{', css) # no whitespace until start of selector
    css = re.sub(r'\s}','}', css) # no whitespace until end of selector
    css = re.sub(r'{\s+', '{', css) # whatspace after {
    css = re.sub(r':\s+',':', css) # no whitespace after colon
    css = re.sub(r';\s+',';', css) # no whitespace in '#fff;  color...'
    css = re.sub(r',\s',r',', css) # no extraspace between commas
    css = re.sub(r'\s+</',r'</', css, re.MULTILINE) # no extraspace before </style>
    css = re.sub(r'}\s+(\w|#)', r'}\1', css)
#    css = re.sub(r'}\s+(\w)', r'}\1', css)
    css = re.sub(r';}', r';}\n', css) # ok to have a linebreak here
    css = simplifyHexColours(css)

    return css.strip()

def _xhtml_slimmer(xhtml):
    # optimize inline CSS
    return _html_slimmer(xhtml, unquote_attributes=0)



def _html_slimmer(html, unquote_attributes=1):
    """ Optimize like XHTML but go one step further """
    # 1. optimize inline CSS
    for styletag in html_inline_css.findall(html):
	html = html.replace(styletag, css_slimmer(styletag))
	
    # 2. Remove excessive whitespace between tags
    html = re.sub(r'>\s+<','><', html)
    
    # 3. Remove oneline comments
    html = html_comments_oneline.sub('', html)
    
    # 4. In every tag, remove quotes on numerical attributes and all
    # excessive whitespace
    for tag in any_tag.findall(html):
	if tag.startswith('<!'):
	    continue
	original = tag
	distinct_matches=[]
	if unquote_attributes:
	    quoted_attributes = quoted_numerical_attribute.findall(tag)
	else:
	    quoted_attributes = []
	for e in quoted_attributes:
	    if e[0] not in distinct_matches:
		# keep track like this to avoid unnecessary duplicate replaces
		distinct_matches.append(e[0])
		tag = tag.replace(e[0], e[0].replace(e[1],''))
		
	# remove >1 repeating whitespace
	tag = excess_whitespace.sub(' ', tag)
	
	# has the tag been improved?
	if original != tag:
	    html = html.replace(original, tag)
    
    
    return html.strip()


## ----- Some fancier names
##

def css_slimmer(css):
    return _css_slimmer(css)

def xhtml_slimmer(xhtml):
    return _xhtml_slimmer(xhtml)

def html_slimmer(html):
    return _html_slimmer(html)


## ----- Methods related to simplifying HEX colour codes

def uniqify(all):
    """ borrowed from Tim Peters' algorithm on ASPN Cookbook """
    u = {}
    for each in all:
        u[each]=1
    return u.keys()

def simplifyHexColours(text):
    """ Replace all colour declarations where pairs repeat.
    I.e. #FFFFFF => #FFF; #CCEEFF => #CEF
    and #EFEFEF, #EFCDI9 avoided """
    colour_replacements = {}
    all_hex_encodings = hex_colour.findall(text)

    for e in uniqify(all_hex_encodings):
        if e[1]==e[2] and e[3]==e[4] and e[5]==e[6]:
            colour_replacements[e] = '#'+e[1]+e[3]+e[5]
    mreplacer = MultiReplacer(colour_replacements)
    return mreplacer(text)

class MultiReplacer:
    def __init__(self, replacements, delimiter='\t', wholeWords=None, caseInsensitive=None):
        # Build replacements dictionary - may come in as a mapping or as a file         
        self.replacements = {}
        try:
            # replacements is a mapping
            self.replacements.update(replacements)
        except TypeError:
            # replacements is a file
            replacementsFile = open(replacements, 'r')
            for line in replacementsFile.readlines():
                fromValue, toValue = line.split(delimiter)[:2] # Split line
                
                while toValue[-1] in '\r\n': # Strip newlines
                    toValue = toValue[:-1]

                self.replacements[fromValue] = toValue # Add to dictionary
            replacementsFile.close()
  
        # Build char to char mapping...
        self.charMap = None
        if not wholeWords:
            charMap = map(chr, range(256))
            for fromValue, toValue in self.replacements.items():
                if len(fromValue) <> 1 or len(toValue) <> 1:
                    break
                if caseInsensitive:
                    charMap[ord(fromValue.upper())] = toValue
                    charMap[ord(fromValue.lower())] = toValue
                else:
                    charMap[ord(fromValue)] = toValue
            else:
                self.charMap = "".join(charMap)
                return

        # String to string mapping - use a regular expression
        import re
        fromVals = self.replacements.keys()
        fromVals.sort()

        # Build regexp pattern
        if not wholeWords:
            rePattern = '|'.join(map(re.escape, fromVals))
        else:
            rePattern = r'\b(' \
                      + '|'.join(map(re.escape, fromVals)) + r')\b'
        
        # Compile regexp
        if caseInsensitive: 
            self.reObject = re.compile(rePattern, re.I)
        else:
            self.reObject = re.compile(rePattern)

    def __call__(self, string):
        # apply replacement to string
        
        # Char to char mapping
        if self.charMap: 
            return string.translate(self.charMap)

        # String to string mapping        
        return self.reObject.sub(self.__replaceMatch, string)
    
    def __replaceMatch(self, match):
        item = match.group(0)
        return self.replacements.get(item)
